@@ -9,6 +9,7 @@ import numpy.core.umath_tests
 import logging
 import datetime
 import pathlib
+import functools
 
 import scipy
 import scipy.stats
@@ -168,7 +169,7 @@ class CollocatedDataset(dataset.HomemadeDataset):
 #                logging.debug("New primary: {!s}".format(gran_prim))
                 try:
                     newprim = self.primary.read(gran_prim)
-                except (ValueError, IOError) as msg:
+                except (dataset.InvalidFileError) as msg:
                     logging.error("Could not read {}: {}".format(
                         gran_prim, msg.args[0]))
                     #continue
@@ -283,8 +284,12 @@ class CollocatedDataset(dataset.HomemadeDataset):
         This is a generator.  It yields in chunks of 100 MB or so.  This
         is so subsequent processing doesn't hog memory needlessly.
 
-        :param datetime start_date: First date to collocate
-        :param datetime end_date: Last date to collocate
+        :param datetime start_date: First date to collocate.  If not
+            given, use last of (self.primary.start_date,
+            self.secondary.start_date).
+        :param datetime end_date: Last date to collocate.  If not given,
+            use earliest of (self.primary.end_date,
+            self.secondary.end_date).
         :param store_size: Store results on disk after passing this size.
             Negative number means no storing.  Size in megabytes.
         :param (list, list) fields: If set, return only those fields.
@@ -307,6 +312,14 @@ class CollocatedDataset(dataset.HomemadeDataset):
 #                ran_prim, gran_sec))
 #            prim = self.primary.read(gran_prim)
 #            sec = self.secondary.read(gran_sec)
+        if start_date is None:
+            start_date = max(self.primary.start_date,
+                self.secondary.start_date)
+
+        if end_date is None:
+            end_date = min(self.primary.end_date,
+                self.secondary.end_date)
+
         logging.info(("Getting collocations {!s} vs. {!s}, distance {:.1f} km, "
                       "interval {!s}, period {!s} - {!s}").format(
                         self.primary.name, self.secondary.name,
@@ -395,6 +408,14 @@ class CollocatedDataset(dataset.HomemadeDataset):
                         numpy.empty(dtype=sec.dtype, shape=0),
                         f)
             else:
+                # select only those within the requested date-range
+                inrange = ((collocs_stored[0]["time"] > start_date) &
+                           (collocs_stored[1]["time"] > start_date) & 
+                           (collocs_stored[0]["time"] < end_date) &
+                           (collocs_stored[1]["time"] < end_date))
+                collocs_stored = (
+                    collocs_stored[0][inrange],
+                    collocs_stored[1][inrange])
                 if fields == (None, None):
                     yield collocs_stored
                 else:
@@ -696,7 +717,8 @@ class CollocationDescriber:
     CollocatedDataset.collocate
 
     target      When smoothing profiles, smooth to this target.  It means
-                we use the averaging kernel and z-grid therefrom.
+                we use the averaging kernel and z-grid therefrom.  The
+                'target' is thus the low resolution profile.
     visualisation  Contains visualisation hints for drawing maps and so.
 
     """
@@ -726,6 +748,28 @@ class CollocationDescriber:
             linestyle="--",
             linewidth=1,
             color="black"))
+
+    figname_compare_profiles = ("compare_profiles_ch4"
+        "_{self.cd.primary.__class__.__name__!s}"
+        "_{self.cd.secondary.__class__.__name__}"
+        "_targ{self.target}"
+        "_{allmask}"
+        "_{quantity}.")
+
+    figname_compare_pc = ("compare_partial_columns_ch4"
+        "_{self.cd.primary.__class__.__name__}"
+        "_{self.cd.secondary.__class__.__name__}"
+        "_targ{self.target}.")
+
+
+    # When removing a priori extending above the averaging kernels, or
+    # high-res profiles extending beyond the averaging kernels, remove all
+    # levels where the fraction of profiles for which this level is too
+    # high is larger than this.
+    toohigh_fraction_cutoff = 0.5
+
+
+            
     _target_vals = ["primary", "secondary"]
 
     _target = None
@@ -758,8 +802,11 @@ class CollocationDescriber:
         self.mask_label = "all"
 
     def filter(self, limit_str="UNDESCRIBED", dist=(0, numpy.inf),
-            interval=(-numpy.inf, numpy.inf),
-            prim_lims={}, sec_lims={}):
+            interval=
+                (-numpy.timedelta64(numpy.iinfo('int64').max, 's'),
+                 +numpy.timedelta64(numpy.iinfo('int64').max, 's')),
+            prim_lims={}, sec_lims={},
+            mask=None):
         """Set limits for further processing.
 
         This method sets a mask that characterises which collocations meet
@@ -774,8 +821,12 @@ class CollocationDescriber:
         :param dict prim_lims: Dictionary with keys corresponding to
             fields in the primary and values (min, max) thereof
         :param dict sec_lims: Like prim_lims but for secondary
+        :param ndarray mask: Alternately, set mask explicitly.  This may
+            be useful for more complicated cases.  Note that the rest of
+            the criteria are still added after.
         """
-        mask = numpy.ones(shape=self.p_col.shape, dtype="bool")
+        if mask is None:
+            mask = numpy.ones(shape=self.p_col.shape, dtype="bool")
 
         (_, _, dists) = self.cd.ellipsoid.inv(
             self.p_col["lon"], self.p_col["lat"],
@@ -1070,13 +1121,17 @@ class CollocationDescriber:
 #                
 #        self.s_col
 
-    def extend_common_grid(self):
-        """Extend both profiles to common grid.
+    def regrid_profiles(self):
+        """Interpolate both profiles to grid of target.
 
-        Sets self.z_grid accordingly.
+        Take the z-grid from the target and interpolate all profiles from
+        primary and secondary to this grid.  If each profile for the
+        target has the same z-grid, the target interpolation will be a
+        no-op.  If the z-grid for the target is not unique, take the mean
+        and grid all profiles to this one.
 
-        If target does not have unique z-grid, choose average and
-        interpolate all onto that.
+        Set self.z_grid to the z_grid that is finally used, and return
+        interpolated profiles (primary, secondary).
         """
 
         targ = (self.p_col, self.s_col)[self.target]
@@ -1103,13 +1158,17 @@ class CollocationDescriber:
 #        for i in self.mask.nonzero()[0]:#range(self.p_col.size):
 #            
 
-    def smooth(self):
+    # This fails because it uses state information that changes between
+    # calls, and this state information is not taken into account.
+    #@functools.lru_cache(maxsize=10)
+    _p_smooth = _s_smooth = None
+    def smooth(self, reload=False):
         """Smooth one profile with the others AK
         
         Normally the profile with the largest information content should
         be smoothed according to the AK of the other.  This is not
-        determined automatically but determined by self.target which is
-        set upon object creation.
+        determined automatically but determined by self.target which
+        should set upon object creation.
 
         Source is equation 4 in:
         
@@ -1117,144 +1176,148 @@ class CollocationDescriber:
         remote sounding instruments.  In: Journal of Geophysical Research,
         Vol 108, No. D3, 4116, doi:10.1029/2002JD002299
 
-        Also calls self.extend_common_grid thus setting self.z_grid.
+        Also calls self.regrid_profiles thus setting self.z_grid.
 
         Returns primary and secondary profiles, one smoothed, the other
-        unchanged
+        unchanged.
+
+        TODO:
+            - Verify the regridding of the a-priori (xa).  It seems that
+              z_xa is sometimes severely negative?  Is this still true?
+
+            - Verify that the regridding of averaging kernels is now
+              correct for profiles where lowest level of low-res
+              AK is above the common z-grid to which we are trying
+              to interpolate, effectively extrapolating, which is
+              ill-conditioned.
+
+            - For calculating partial columns, any regridding needs to be
+              also applied to (at least) T and P!  Better set p_smooth,
+              s_smooth as ndarrays containing several types...
+
+            - Should turn around AK axes in reading routine, not here!
+              And/or only if really needed!
+
+            - Clean up the code... should first decide on a z-grid and
+              then regrid everything (raw, smoothed, p, T, etc.) onto
+              this?  And split in smaller methods/functions, this method
+              is too large!
         """
 
+        # Cache results "by hand"; it appears lru_cache doesn't work well
+        # with methods depending on the state of a (mutable) object
+        if not reload and (self._p_smooth is not None) and (self._s_smooth is not None):
+            return (self._p_smooth, self._s_smooth)
+
         targ = (self.p_col, self.s_col)[self.target][self.mask]
         targobj = (self.cd.primary, self.cd.secondary)[self.target]
-        #highres = (self.prim, self.sec)[1-self.target]
-        (p_ch4_int, s_ch4_int) = self.extend_common_grid()
 
-        extra = xa = p_xa = z_xa = ak = p_ak = z_ak = None
-        if not "CH4_apriori" in targ.dtype.names: # try to add
-            extra = targobj.get_additional_field(targ, "(smoothing)")
-            xa = extra["ch4_ap"]
-            p_xa = extra["p_ch4_ap"]
-            if "z_ch4_ap" in extra.dtype.names:
-                z_xa = extra["z_ch4_ap"]
-            #xa = targobj.get_additional_field(targ, "CH4_apriori")
-        else:
-            xa = targ["CH4_apriori"]
-            p_xa = targ["p"]
-            z_xa = targ["z"]
+        # Regrid primary and secondary CH4 profiles so that they share the
+        # same z_grid, that will be set to self.z_grid.
+        #
+        # NB: if a profile's z_grid does not extend to the full range of
+        # self.z_grid, any "extrapolated" values are set to numpy.nan by
+        # regrid_profiles (through scipy.interpolate.interp1d)
+        (p_ch4_int, s_ch4_int) = self.regrid_profiles()
+
+        # Both p_ch4_int and s_ch4_int are
+        # already regridded to both be on self.z_grid (see above)
+
         xh = (p_ch4_int, s_ch4_int)[1-self.target]
+        z_xh = self.z_grid
 
-        if "CH4_ak" in targ.dtype.names:
-            ak = targ["CH4_ak"]
-            p_ak = p_xa
-            z_ak = z_xa
-        elif extra is None:
-            ak = targobj.get_additional_field(targ, "CH4_ak")
-            raise RuntimeError("Found a priori but not AK outside ?!")
-        else:
-            ak = extra["ch4_ak"]
-            p_ak = extra["p_ch4_ak"]
-            if "z_ch4_ak" in extra.dtype.names:
-                z_ak = extra["z_ch4_ap"]
+        # Get "pure" a priori and averaging kernels.  May or may not be on
+        # the same grid as target objects.
+        (xa, z_xa, ak, z_ak) = self._get_xa_ak(targ, targobj)
 
-        # axes tend to be turned around?  FIXME VERIFY
-        # Should rather do this in reading routine?!
-        A = ak.swapaxes(1, 2)
+        # FIXME: Should rather do this in reading routine...?
+        # but: needs a priori that may be available only "elsewhere"
+        # FIXME: Separate conversion and swapping
+        if targobj.A_needs_converting:
+            # correct ak according to e-mail Stephanie 2014-06-17
+            ak = ak.swapaxes(1, 2)
+            ak = numpy.rollaxis(
+                numpy.dstack(
+                    [pamath.convert_ak_ap2vmr(
+                        ak[i, :, :], xa[i, :])
+                    for i in range(xa.shape[0])]), 2, 0)
 
-        # I may need to regrid xh and ak to the resolution of targ
-#        if not (ak.shape[1] == ak.shape[2] == xa.shape[1]):
-#            raise ValueError("A and xa shapes don't match!")
+        ## Make sure everything is on the same grid:
+        #
+        # - Actual profiles
+        #       -> regrid_profiles(), called above, guarantees those are
+        #       on the same grid, set in self.z_grid.  This should be from
+        #       the low-resolution profile (i.e. the target).
+        #       Note that z_xh == self.z_grid
+        # - Averaging kernels for low-resolution profile
+        #       -> currently on z_ak
+        # - A priori for low-resolution profile, on z_xa
+        #       -> currently on z_xa
+        #
+        # It is useless if any of those extends above any of the others,
+        # so we choose the lowest maximum z and cut off everything to
+        # that.
 
-        if z_xa is None: # convert from p_xa
-            if (p_xa.shape == targ["p"].shape) and (p_xa - targ["p"]).max() < 1e-3: # same grid
-                z_xa = targ["z"]
-            elif (p_xa.shape == targ["T"].shape == targ["h2o"].shape):
-                # cannot be vectorised :(
-                filler = numpy.empty(shape=(p_xa.shape[1],), dtype="f4")
-                filler.fill(numpy.nan)
-                z_xa = [(physics.p2z_hydrostatic(
-                            p_xa[i, :],
-                            targ[i]["T"],
-                            targ[i]["h2o"],
-                            targ[i]["p0"],
-                            targ[i]["z0"],
-                            targ[i]["lat"],
-                            -1, extend=True)
-                                if numpy.isfinite(xa[i]).any() else filler)
-                            for i in range(targ.shape[0])]
-                z_xa = numpy.vstack(z_xa)
-            else:
-                raise ValueError("Can't find z.  Should I try harder?")
+        (xa, z_xa, ak, z_ak, xh, z_xh, p_ch4_int, s_ch4_int) = \
+            self._limit_to_shared_range(xa, z_xa, ak, z_ak, xh, z_xh, 
+                p_ch4_int, s_ch4_int)
 
-        if z_ak is None: # convert from p_ak
-            if (p_ak.shape == targ["p"].shape) and numpy.nanmax(p_ak - targ["p"]) < 1e-3:
-                z_ak = targ["z"]
-            elif (p_ak.shape == p_xa.shape) and (p_ak - p_xa).max() < 1e-3:
-                z_ak = z_xa
-            else:
-                raise ValueError("Don't know how to get z.  Should I try harder?")
-
-        if z_ak.shape != xh.shape:
-            # regrid from z_xa to z_ak
-            z_ak_new = targ["z"]
-            z_ak_old = z_ak
-            ak_old = ak
-            W = numpy.dstack([
-                pamath.linear_interpolation_matrix(
-                    z_ak_old[i, :], z_ak_new[i, :])
-                        for i in range(z_ak.shape[0])])
-            W = numpy.transpose(W, [2, 0, 1])
-#            z_ak_new = numpy.vstack([W[i, :, :].dot(z_ak[i, :]) for i in
-#                        range(W.shape[0])])
-            ak_new = numpy.dstack(
-                [pamath.apply_W_A(W[i, :, :], ak_old[i, :, :])
-                    for i in range(W.shape[0])])
-            ak = ak_new
-
-        if z_xa.shape != xh.shape:
-            z_xa_new = targ["z"]
-            z_xa_old = z_xa
-            xa_old = xa
-            W = numpy.dstack([
-                pamath.linear_interpolation_matrix(
-                    z_xa_old[i, :], z_xa_new[i, :])
-                        for i in range(z_xa.shape[0])])
-            W = numpy.transpose(W, [2, 0, 1])
-            xa_new = numpy.vstack([
-                W[i, :, :].T.dot(xa_old[i, :]) for i in range(W.shape[0])])
-            xa = xa_new
+        (xa, z_xa, ak, z_ak) = self._regrid_xa_ak(xa, z_xa, ak, z_ak, z_xh)
 
 
-        # OK :)
+        # OK :).  Now everything should be on the same z-grid!
+        #
+        # Prepare the ndarrays to populate.
+
+        (nprof, nlev) = p_ch4_int.shape
+        p = numpy.zeros(shape=nprof,
+            dtype=[("CH4", "f4", nlev),
+                   ("p", "f4", nlev),
+                   ("T", "f4", nlev),
+                   ("z", "f4", nlev)])
+        s = numpy.zeros_like(p)
+
 
         # where high-res profile is outside its z-range, set to a-priori
         # of low-res.  For example, PEARL may go down only to 19.5 km, but
         # the low-res xa might go down to 5 km.  Then set [5, 19.5] of
         # high-res equal to a priori of other.
+
         xh[numpy.isnan(xh)] = xa[numpy.isnan(xh)]
 
-        if targobj.A_needs_converting:
-            # correct A according to e-mail Stephanie 2014-06-17
-            A = numpy.rollaxis(
-                numpy.dstack(
-                    [numpy.diag(1/targ["CH4_apriori"][i, :]).dot(
-                        A[i, :, :]).dot(
-                        numpy.diag(targ["CH4_apriori"][i, :]))
-                    for i in range(targ.shape[0])]), 2, 0)
-
-        #xs = xa + A.dot(xh-xa)
-        xb = xh - xa
+        # This is where the smoothing is actually performed!
         xs = numpy.vstack(
-            [xa[n, :] + A[n, :, :].dot(xb[n, :]) for n in range(A.shape[0])])
-#        xs = xa + numpy.core.umath_tests.matrix_multiply(A, xb[..., numpy.newaxis]).squeeze()
+            [pamath.smooth_profile(xh[n, :], ak[n, :, :], xa[n, :])
+                for n in range(ak.shape[0])])
 
         # remove invalid data
         OK = numpy.isfinite(xs).any(1)
+        # I don't really know what to do with still incomplete averaging
+        # kernels (i.e. flagged levels), as I want all profiles on the
+        # same grid.  Remove those for now.
+        invalid = (ak<-10).any(1).any(1)
+        OK = OK & (~invalid)
+
         xs = xs[OK, :]
         p_ch4_int = p_ch4_int[OK, :]
         s_ch4_int = s_ch4_int[OK, :]
 
+        if not all(numpy.array_equal(self.z_grid, z) for z in
+            (z_xa, z_ak, z_xh)):
+            raise RuntimeError("z_grids should be equal by now!")
+
+        self.z_smooth = z_xa
         if self.target == 0:
+            p["CH4"] = p_ch4_int
+            s["CH4"] = xs
+            self._p_smooth = p_ch4_int
+            self._s_smooth = xs
             return (p_ch4_int, xs)
         elif self.target == 1:
+            p["CH4"] = xs
+            s["CH4"] = s_ch4_int
+            self._p_smooth = xs
+            self._s_smooth = s_ch4_int
             return (xs, s_ch4_int)
         else:
             raise RuntimeError("Impossible!")
@@ -1274,7 +1337,8 @@ class CollocationDescriber:
 
         #return numpy.array([
         return {k:
-            numpy.array([scipy.stats.scoreatpercentile(D[k][:, i], percs)
+            numpy.array([scipy.stats.scoreatpercentile(
+                D[k][numpy.isfinite(D[k][:, i]), i], percs)
                 for i in range(D[k].shape[1])])
             for k in D.keys()}
 
@@ -1310,6 +1374,85 @@ class CollocationDescriber:
 
         return self._compare_profiles(p_ch4_int, s_ch4_int, percs=percs)
 
+    def plot_aks(self):
+        """Visualise averaging kernels.
+
+        Will average all averaging kernels and plot them as lines.
+        """
+
+        if "ak" in self.cd.primary.aliases:
+            # make a copy because I'm going to set flagged values to 'nan'
+            p_ak = self.p_col[self.cd.primary.aliases["ak"]].copy()
+        else:
+            p_ak = None
+
+        if "ak" in self.cd.secondary.aliases:
+            s_ak = self.s_col[self.cd.secondary.aliases["ak"]].copy()
+        else:
+            s_ak = None
+
+        f = matplotlib.pyplot.figure()
+        a1 = f.add_subplot(1, 2, 1)
+        a2 = f.add_subplot(1, 2, 2)
+        a_both = []
+
+        mx = mn = 0
+        data = []
+        if "ak" in self.cd.primary.aliases:
+            p_ak[p_ak<-100] = numpy.nan # presumed flagged
+            mean_p_ak = numpy.nanmean(p_ak, 0)
+            a1.plot(mean_p_ak.T, self.p_col["z"].mean(0))
+            a1.set_title(self.cd.primary.name)
+            a_both.append(a1)
+            mx = numpy.nanmax(mean_p_ak)
+            mn = numpy.nanmin(mean_p_ak)
+            data.append(numpy.hstack(
+                (self.p_col["z"].mean(0)[:, numpy.newaxis], mean_p_ak)))
+
+        if "ak" in self.cd.secondary.aliases:
+            s_ak[s_ak<-100] = numpy.nan # presumed flagged
+            mean_s_ak = numpy.nanmean(s_ak, 0)
+            a2.plot(mean_s_ak.T, self.s_col["z"].mean(0))
+            a2.set_title(self.cd.secondary.name)
+            a_both.append(a2)
+            mx = numpy.max([mx, numpy.nanmax(mean_s_ak)])
+            mn = numpy.min([mn, numpy.nanmin(mean_s_ak)])
+            data.append(numpy.hstack(
+                (self.s_col["z"].mean(0)[:, numpy.newaxis], mean_s_ak)))
+
+        for a in a_both:
+            a.set_xlabel("Mean averaging kernel []")
+            a.set_xlim([mn, mx])
+            a.set_ylim([0, 40e3])
+
+        graphics.print_or_show(f, False,
+            "ak_{}_{}.".format(self.cd.primary.__class__.__name__,
+                               self.cd.secondary.__class__.__name__),
+            data=data)
+
+        # And the matrices
+
+        
+        logging.info("Summarising sensitivities")
+        if p_ak is not None:
+            paks = physics.AKStats(p_ak, 
+                name="{}_from_{}".format(
+                    self.cd.primary.__class__.__name__,
+                    self.cd.secondary.__class__.__name__))
+            with numpy.errstate(invalid="warn"):
+                paks.plot_sensitivity_range(z=self.p_col["z"])
+                #paks.plot_sensitivity_density(z=numpy.nanmean(self.p_col["z"], 0))
+                paks.plot_sensitivity_density(z=self.p_col["z"])
+
+        if s_ak is not None:
+            saks = physics.AKStats(s_ak, 
+            name="{}_from_{}".format(
+                self.cd.secondary.__class__.__name__,
+                self.cd.primary.__class__.__name__))
+            with numpy.errstate(invalid="warn"):
+                saks.plot_sensitivity_range(z=self.s_col["z"])
+                #saks.plot_sensitivity_density(z=numpy.nanmean(self.s_col["z"], 0))
+                saks.plot_sensitivity_density(z=self.s_col["z"])
 
     def visualise_profile_comparison(self, z_grid, filters=None):
         """Visualise profile comparisons.
@@ -1331,6 +1474,7 @@ class CollocationDescriber:
 
         colours = {}
         percs = {}
+        profs = {}
         lims = {}
         xlabels = dict(
             diff = "Delta CH4 [ppv]",
@@ -1339,19 +1483,24 @@ class CollocationDescriber:
             prim = "Primary CH4 [ppv]",
             sec = "Secondary CH4 [ppv]")
         xlims = dict(
-            diff = (-3e-7, 3e-7),
-            rmsd = (0e-6, 5e-7),
-            ratio = (0.8, 1.2),
+            diff = (-1e-7, 1e-7),
+            rmsd = (0, 2e-7),
+            ratio = (0.9, 1.9),
             prim = (0, 2e-6),
             sec = (0, 2e-6))
         self.reset_filter()
 
         # see how smoothed or raw compare
 
+        filter_modes = []
+
+        filter_modes.append(self.mask_label)
         z_grids = {}
+#        profs[self.mask_label + "_smooth"] = self.smooth()
         percs[self.mask_label + "_smooth"] = self.compare_profiles_smooth(
                     z_grid, p_locs)
         z_grids["smooth"] = self.z_grid
+#        profs[self.mask_label + "_raw"] = self.interpolate_profiles(z_grid)
         percs[self.mask_label + "_raw"] = self.compare_profiles_raw(
                     z_grid, p_locs)
         z_grids["raw"] = z_grid
@@ -1364,23 +1513,24 @@ class CollocationDescriber:
             self.filter(**fd)
             percs[fd["limit_str"] + "_raw"] = self.compare_profiles_raw(z_grid, p_locs)
             percs[fd["limit_str"] + "_smooth"] = self.compare_profiles_smooth(z_grid, p_locs)
+            filter_modes.append(self.mask_label)
 #        percs = self.compare_profiles(z_grid, p_locs)
         #for (i, v) in enumerate("diff diff^2 ratio".split()):
+        # quantities such as diff, ratio, rmsd
         for quantity in percs[self.mask_label + "_raw"].keys():
             f = matplotlib.pyplot.figure()
             a = f.add_subplot(1, 1, 1)
             data = dict(smooth=[], raw=[])
-            for filt in percs.keys():
-                ff = "smooth" if "smooth" in filt else "raw"
-                #if "smooth" in filt: z_grid = z_grids["smooth"]
-                #elif "raw" in filt: z_grid = z_grids["raw"]
-                #else: raise RuntimeError("Something wrong")
-                for k in range(len(p_locs)):
-                    a.plot(percs[filt][quantity][:, k], z_grids[ff],
-                           color=colours[filt],
-                           linestyle=p_styles[k], linewidth=p_widths[k],
-                           label=(filt if k==2 else None))
-                    data[ff].append(percs[filt][quantity][:, k])
+            # filters such as all, nearby, small delta-SPV
+            for filter_mode in filter_modes: # percs.keys():
+                for ff in ("smooth", "raw"):
+                    filt = "{}_{}".format(filter_mode, ff)
+                    for k in range(len(p_locs)):
+                        a.plot(percs[filt][quantity][:, k], z_grids[ff],
+                               color=colours[filt],
+                               linestyle=p_styles[k], linewidth=p_widths[k],
+                               label=(filt if k==2 else None))
+                        data[ff].append(percs[filt][quantity][:, k])
             # end for percentiles
             # end for filters
 #            a.plot(percs[1], z_grid, label="p diff^2", color="red")
@@ -1394,20 +1544,44 @@ class CollocationDescriber:
                 "CH4 {}, {} vs. {}".format(quantity,
                     self.cd.primary.name, self.cd.secondary.name))
             a.grid(which="major")
-            a.set_ylim([5e3, 60e3])
-            ### FIXME: write data for both, will need to be in two
-            ### files...
+            a.set_ylim([5e3, 50e3])
+            # Set y-lim according to ak's
+            # (never mind for now, will set by hand in LaTeX code)
+#            p_ak = self.p_col[self.cd.primary.aliases["ak"]]
+#            s_ak = self.s_col[self.cd.secondary.aliases["ak"]]
+
+            a.text(xlims[quantity][0]+0.1*xlims[quantity][1], 45e3, "{:d} profiles".format(self.mask.sum()))
+            allmask = ','.join(filter_modes)
             graphics.print_or_show(f, False,
-                "compare_profiles_ch4_{}_{}_{}_{}_{}_{}.".format(
-                    self.cd.primary.__class__.__name__,
-                    self.cd.primary.name.replace(" ", "_"),
-                    self.cd.secondary.__class__.__name__,
-                    self.cd.secondary.name.replace(" ", "_"),
-                    quantity,
-                    "multi"),
-                    data=numpy.vstack((z_grids["raw"],)+tuple(data["raw"])).T)
+                self.figname_compare_profiles.format(**vars()),
+                data=
+                    (numpy.vstack((z_grids["raw"],)+tuple(data["raw"])).T,
+                     numpy.vstack((z_grids["smooth"],)+tuple(data["smooth"])).T)
+                    )
             matplotlib.pyplot.close(f)
         # end for quantities
+
+        ## Plot all profiles
+##        f = matplotlib.pyplot.figure()
+##        a = f.add_subplot(1, 1, 1)
+##        for (w, c) in [("raw", "black"), ("smooth", "red")]:
+##            a.plot(profs[self.mask_label + "_" + w][self.target].T, z_grids[w],
+##                  color=c, label=w + " prim", linewidth=0.3)
+###            a.plot(profs[self.mask_label + "_" + w][1].T, z_grids[w],
+###                  color=c, label=w + " sec", linewidth=0.3)
+###        a.legend()
+##        a.set_xlabel("CH4 [ppv]")
+##        a.set_ylabel("Elevation [m]")
+##        a.set_title("All profiles")
+##        a.grid(which="major")
+##        a.set_ylim([5e3, 60e3])
+##        a.set_xlim([0, 2e-6])
+##        graphics.print_or_show(f, False,
+##            self.figname_compare_profiles.format(self=self,
+##                quantity="spaghetti"))
+##        matplotlib.pyplot.close(f)
+        #
+
         self.reset_filter()
 
         # some specialised plots
@@ -1435,6 +1609,218 @@ class CollocationDescriber:
 ##                     (z_grid, iqr["prim"], iqr["sec"], iqr["diff"])).T)
 ##         matplotlib.pyplot.close(f)
 
+    def partial_columns(self, smoothed=True):
+        """Calculate partial columns.
+
+        """
+
+        shared_range = (max(self.cd.primary.range[0],
+                            self.cd.secondary.range[0]),
+                        min(self.cd.primary.range[1],
+                            self.cd.secondary.range[1]))
+
+        if smoothed:
+            (p, s) = self.smooth()
+            z = self.z_smooth
+        else:
+            raise NotImplementedError()
+
+        # levels in 'z' within shared_range
+        valid_range = (z > shared_range[0]) & (z < shared_range[1])
+
+        z_valid = z[valid_range]
+        p_valid_vmr = p.T[valid_range, :]
+        s_valid_vmr = s.T[valid_range, :]
+
+        p_valid_nd = physics.vmr2nd(p_valid_vmr, p_T, p_p)
+        s_valid_nd = physics.vmr2nd(s_valid_vmr, s_T, s_p)
+
+        p_parcol = pamath.integrate_with_height(
+            z_valid, p_valid_nd)
+        s_parcol = pamath.integrate_with_height(
+            z_valid, s_valid_nd)
+
+        return (p_parcol, s_parcol, (z[valid_range].min(), z[valid_range].max()))
+
+    def visualise_pc_comparison(self):
+        """Visualise comparison for partial columns
+        """
+        (p_parcol, s_parcol, valid_range) = self.partial_columns(smoothed=True)
+
+        (f, a) = matplotlib.pyplot.subplots()
+        valid = numpy.isfinite(p_parcol) & numpy.isfinite(s_parcol)
+        a.plot(p_parcol[valid], s_parcol[valid], '.')
+        mx = max(p_parcol[valid].max(), s_parcol[valid].max())
+        mn = min(p_parcol[valid].min(), s_parcol[valid].min())
+        a.plot([0, 2*mx], [0, 2*mx], linewidth=2, color="black")
+        a.set_xlim(0.9*mn, 1.1*mx)
+        a.set_ylim(0.9*mn, 1.1*mx)
+        a.set_xlabel("CH4 {:s} [unit]".format(self.cd.primary.name))
+        a.set_ylabel("CH4 {:s} [unit]".format(self.cd.secondary.name))
+        a.set_title(("Partial columns {:.1f}--{:.1f}, "
+                     "difference {:s}-{:s}").format(
+                        valid_range[0]/1e3, valid_range[1]/1e3,
+                        self.cd.primary.name, self.cd.secondary.name))
+        graphics.print_or_show(f, None, self.figname_compare_pc.format(**vars()))
+
+    def _get_xa_ak(self, targ, targobj):
+        """Helper for smooth(...)
+        """
+
+        z_xa = z_ak = None
+
+        ## Look for a priori and z-grid (will need to regrid later)
+        if "ap" in targobj.aliases:
+            xa = targ[targobj.aliases["ap"]]
+            p_xa = targ["p"]
+            z_xa = targ["z"]
+        else:
+            extra = targobj.get_additional_field(targ, "(smoothing)")
+            xa = extra["ch4_ap"]
+            p_xa = extra["p_ch4_ap"]
+            if "z_ch4_ap" in extra.dtype.names:
+                z_xa = extra["z_ch4_ap"]
+            #xa = targobj.get_additional_field(targ, "CH4_apriori")
+
+        ## Look for averaging kernel and z-grid (will need to regrid later)
+        if "ak" in targobj.aliases:
+            ak = targ[targobj.aliases["ak"]]
+            p_ak = targ["p"]
+            z_ak = targ["z"]
+        elif extra is None:
+            ak = targobj.get_additional_field(targ, "CH4_ak")
+            raise RuntimeError("Found a priori but not AK outside ?!")
+        else:
+            ak = extra["ch4_ak"]
+            p_ak = extra["p_ch4_ak"]
+            if "z_ch4_ak" in extra.dtype.names:
+                z_ak = extra["z_ch4_ap"]
+
+        if z_xa is None: # convert from p_xa
+            if (p_xa.shape == targ["p"].shape) and (p_xa - targ["p"]).max() < 1e-3: # same grid
+                z_xa = targ["z"]
+            elif (p_xa.shape == targ["T"].shape == targ["h2o"].shape):
+                # cannot be vectorised :(
+                filler = numpy.empty(shape=(p_xa.shape[1],), dtype="f4")
+                filler.fill(numpy.nan)
+                z_xa = [(physics.p2z_hydrostatic(
+                            p_xa[i, :],
+                            targ[i]["T"],
+                            targ[i]["h2o"],
+                            targ[i]["p0"],
+                            targ[i]["z0"],
+                            targ[i]["lat"],
+                            -1, extend=True)
+                                if numpy.isfinite(xa[i]).any() else filler)
+                            for i in range(targ.shape[0])]
+                z_xa = numpy.vstack(z_xa)
+            else:
+                raise ValueError("Can't find z.  Should I try harder?")
+
+        if z_ak is None: # convert from p_ak
+            if (p_ak.shape == targ["p"].shape) and numpy.nanmax(p_ak - targ["p"]) < 1e-3:
+                z_ak = targ["z"]
+            elif (p_ak.shape == p_xa.shape) and numpy.nanmax(p_ak - p_xa) < 1e-3:
+                z_ak = z_xa
+            else:
+                raise ValueError("Don't know how to get z.  Should I try harder?")
+        return (xa, z_xa, ak, z_ak)
+
+
+    def _limit_to_shared_range(self, xa, z_xa, ak, z_ak, xh, z_xh,
+        p_ch4_int, s_ch4_int):
+        """Helper for smooth(...)
+        """
+        lowest_z_max = min(numpy.nanmax(z_ak), numpy.nanmax(z_xa),
+            z_xh.max())
+
+        if z_xh.max() > lowest_z_max:
+            # cut off actual profiles
+            toohigh = self.z_grid > lowest_z_max
+            xh = xh[:, ~toohigh]
+            z_xh = z_xh[~toohigh]
+            self.z_grid = z_xh
+            # NB: to smooth the high-res measurement `xh` with the
+            # low-resolution averaging kernel `ak` and a priori `xa`, the
+            # low--res measurement is not used.  However, it is still
+            # returned, so should still be cut off so caller can
+            # consistently process. 
+            p_ch4_int = p_ch4_int[:, ~toohigh]
+            s_ch4_int = s_ch4_int[:, ~toohigh]
+
+        if numpy.nanmax(z_ak) > lowest_z_max:
+            # Remove all levels where more than a fraction of the a priori
+            # have no information due to the levels being too high.
+            st = numpy.seterr(invalid="ignore")
+            toohigh = ((z_ak > lowest_z_max).sum(0) / z_ak.shape[0]
+                            > self.toohigh_fraction_cutoff)
+            numpy.seterr(**st)
+            # see http://stackoverflow.com/q/26259662/974555 for a
+            # rationale on this 'roundabout' way
+            ak = ak[:, ~toohigh, :][:, :, ~toohigh]
+            z_ak = z_ak[:, ~toohigh]
+
+        if numpy.nanmax(z_xa) > lowest_z_max:
+            st = numpy.seterr(invalid="ignore")
+            toohigh = ((z_xa > lowest_z_max).sum(0) / z_xa.shape[0]
+                           > self.toohigh_fraction_cutoff)
+            numpy.seterr(**st)
+            xa = xa[:, ~toohigh]
+            z_xa = z_xa[:, ~toohigh]
+
+        return (xa, z_xa, ak, z_ak, xh, z_xh, p_ch4_int, s_ch4_int)
+
+    def _regrid_xa_ak(self, xa, z_xa, ak, z_ak, z_xh):
+        """Helper for smooth(...)
+        """
+        # WARNING: What if z_ak.shape == z_xh.shape
+        # but z_ak != z_xh?
+        if z_ak.shape == z_xh.shape and not tools.array_equal_with_equal_nans(z_ak, z_xh):
+            raise NotImplementedError("Improve z-axis checking!")
+        if z_ak.shape != z_xh.shape:
+            # regrid from z_xa to z_ak
+            #z_ak_new = targ["z"]
+            z_ak_new = z_xh
+            z_ak_old = z_ak
+            ak_old = ak
+            with numpy.errstate(invalid="ignore"):
+                ak_new = numpy.dstack([pamath.regrid_ak(
+                    ak_old[i, :, :], z_ak_old[i, :], z_ak_new, cut=True)
+                        for i in range(z_ak.shape[0])])
+            ak = numpy.rollaxis(ak_new, 2, 0)
+            z_ak = z_ak_new
+
+        # WARNING: What if z_xa.shape == z_xh.shape
+        # but z_xa != z_xh?
+        if z_xa.shape == z_xh.shape and not tools.array_equal_with_equal_nans(z_xa, z_xh):
+            raise NotImplementedError("Improve z-axis checking!")
+        if z_xa.shape != z_xh.shape:
+            #z_xa_new = targ["z"]
+            z_xa_new = z_xh
+            z_xa_old = z_xa
+            xa_old = xa
+            W = numpy.dstack([
+                pamath.linear_interpolation_matrix(
+                    z_xa_old[i, :], z_xa_new)
+                        for i in range(z_xa.shape[0])])
+            W = numpy.transpose(W, [2, 0, 1])
+            xa_new = numpy.vstack([
+                W[i, :, :].T.dot(xa_old[i, :]) for i in range(W.shape[0])])
+            # interpolation above sets values outside the range to the
+            # edge of the range, leading to a weird constant in CH4(z).
+            # We don't want that, set to nan instead.
+            # (disable temporarily so I have SOME result to show for PAHA)
+            # FIXME find a proper solution for this!
+            #ok = numpy.isfinite(xa_new).any(1)
+            #outside = z_xa_new[ok, :] > z_xa_old.max(1)[ok, numpy.newaxis]
+            #xanew_tmp = xa_new[ok, :]
+            #xanew_tmp.flat[outside.ravel()] = numpy.nan
+            #xa_new[ok, :] = xanew_tmp
+            xa = xa_new
+            z_xa = z_xa_new
+
+        return (xa, z_xa, ak, z_ak)
+
 def find_collocation_duplicates(p_col, s_col):
     dt = numpy.dtype([("A", "f8"), ("B", "f8"), ("C", "M8[s]"), ("D",
     "f8"), ("E", "f8"), ("F", "f8")])
