@@ -49,8 +49,14 @@ class TansoFTSv10x(dataset.MultiFileDataset, TansoFTSBase):
     """
 
     re = r"GOSATTFTS(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})_02P02TV010[01]R\d{6}[0-9A-F]{5}\.h5"
-    aliases = {"CH4_profile": "ch4_profile"}
+    aliases = {"CH4_profile": "ch4_profile", "ak": "ch4_ak"}
     n_prof = "p"
+    # For TANSO collocating with PEARL, 50% of profiles have > 40%
+    # sensitivity between 5.3 km and 9.7 km.
+    #
+    # For TANSO collocating with ACE, 20% of profiles have > 30%
+    # sensitivity between 5.2 km and 9.5 km.
+    range = (5e3, 10e3)
 
     # NOTE: For Tanso FTS v1.0x, there are THREE pressure profiles:
     #
@@ -100,6 +106,8 @@ class TansoFTSv10x(dataset.MultiFileDataset, TansoFTSBase):
         # z-grids.  Want z-grid corresponding to CH4-profile.
         p_for_T = self.p_for_T_profile # down up to 1 kPa
         p_for_CH4 = obj["p"] # down to ~50 Pa
+        p_valid = p_for_CH4>0
+        p_for_CH4 = p_for_CH4[p_valid]
         # inter- and extra-polate T and h2o onto the p_for_CH4 grid
         # This introduces an extrapolation error but between 1 kPa and 50
         # Pa I don't really care, should be insignificant.
@@ -183,7 +191,11 @@ class TansoFTSv10x(dataset.MultiFileDataset, TansoFTSBase):
         # original order, but only for those that were part of p_for_CH4
         p_i_was_CH4 = ((p_inds>=p_for_ch4_i0) & (p_inds<p_for_ch4_i1)).nonzero()[0]
 
-        return z_extp[p_i_was_CH4]#.data # no masked array...
+        z_ret = numpy.zeros(dtype="f4", shape=obj["p"].shape)
+        z_ret[p_valid] = z_extp[p_i_was_CH4]
+        z_ret[~p_valid] = numpy.nan
+        return z_ret
+        #return z_extp[p_i_was_CH4]#.data # no masked array...
 #        inv_inds[-len(p_for_CH4):]]
 
 #        z_for_T = physics.p2z_hydrostatic(
@@ -534,8 +546,9 @@ class ACEFTS(dataset.SingleMeasurementPerFileDataset,
     _time_format = "%Y-%m-%d %H:%M:%S"
     aliases = {"CH4_profile": "CH4"}
     filename_fields = {"orbit": "u4", "version": "S3"}
-    unique_fields = {"orbit"}
+    unique_fields = {"orbit", "time"}
     n_prof = "z"
+    range = (5e3, 150e3)
 
     @staticmethod
     def read_header(fp):
@@ -576,7 +589,12 @@ class ACEFTS(dataset.SingleMeasurementPerFileDataset,
                 # why does this not work?
                 # http://stackoverflow.com/q/22865877/974555
                 #D[names][n] = tuple(float(f) for f in line.split())
-                vals = tuple(float(f) for f in line.split())
+                try:
+                    vals = tuple(float(f) for f in line.split())
+                except ValueError:
+                    # raise InvalidFileError instead so I can catch more
+                    # narrowly higher up in the stack
+                    raise dataset.InvalidFileError("Unable to read content")
                 for (i, name) in enumerate(names):
                     D[name][n] = vals[i]
 
