@@ -7,16 +7,266 @@
 Mostly obtained from PyARTS
 """
 
+import logging
 import numbers
 
 import numpy
+import matplotlib
 
 import pyproj
 
+
 #from .constants import (h, k, R_d, R_v, c)
 from . import constants as c
 from . import math as pamath
 from . import tools
+from . import graphics
+
+#import trollimage.colormap
+
+class AKStats:
+
+    filename = "sensitivity_{mode}_matrix_{name}."
+    cmap = "afmhot_r"
+
+    def __init__(self, aks, name="UNDEFINED"):
+        self.aks = aks.copy()
+        with numpy.errstate(invalid="ignore"):
+            self.aks[self.aks<=-999] = numpy.nan
+        self.name = name
+
+    def dofs(self):
+        """Calculate degrees of freedom.
+
+        According to Rodgers (2000), equation (2.80), page 37.
+
+        This is the trace of the averaging kernels.
+        """
+
+        return self.aks.trace(axis1=1, axis2=2)
+
+    def sensitivities(self):
+        """Calculate sensitivities.
+
+        According to:
+          R. L. Batchelor et al.: Ground-based FTS comparisons ad ACE
+          validation at Eureka during IPY.  Page 57.
+        Sum of each row of the averaging kernel matrix defines sensitivity
+        to measurument.  Note that I - A is the sensitivity to the a priori,
+        soand the sum of the rows of (I - A) + the sum of the rows of A
+        equals 1; so these two could be interpreted as percentages.
+        """
+
+        return self.aks.sum(1)
+
+    def sensitivity_density_matrix(self, sens_fractions=numpy.linspace(0, 1, 11)):
+        """What fraction of profiles have sensitivity >x at level y?
+
+        :param ndarray sens_fractions: Fractions x to consider
+        :returns: (sens_fractions, sens_mat),
+            where sens_mat is a matrix with fraction at each level with at
+            least sensitivity x.
+        """
+
+        sensitivities = self.sensitivities()
+        with numpy.errstate(invalid="ignore"):
+            sensmat = numpy.vstack([(sensitivities>x).sum(0) / sensitivities.shape[0]
+                for x in sens_fractions])
+        return (sens_fractions, sensmat)
+
+    def sensitivity_range_matrix(self,
+        sens_fractions = numpy.linspace(0, 1, 11),
+        sens_counters = None):
+        """
+        For each profile, how many layers have sensitivity of at least x?
+        
+        Creates a "sensitivity score" matrix.  How many profiles have at
+        least y layers of sensitivity >= x?
+
+        :param ndarray sens_fractions:
+            Array of shape (N,).
+            Sensitivities to consider.  Defaults to 0, 0.1, ..., 1.0.
+            Will be used to count no. of profiles with sensitivity larger
+            than this fraction.
+        :param ndarray sens_counters:
+            Integer array of shape (p,),
+            indicating the count as to how many profiles
+            have at least this many layers with sensitivity larger than x.
+            Defaults to arange(self.aks.shape[1]+1)
+        :returns:
+            Tuple (sens_fractions, sens_counters, sensmat), where sensmat
+            is an N x p matrix, N being the number of sensitivities to
+            consider, and p the counters.  In each coordinate define by
+            the ararys sens_fractions and sens_counters, it contains the
+            fraction of profiles that have at least p levels above
+            sensitivity x.
+        """
+        
+        if sens_counters is None:
+            sens_counters = numpy.arange(self.aks.shape[1]+1)
+
+        sensitivities = self.sensitivities()
+        with numpy.errstate(invalid="ignore"):
+            sensmat = numpy.vstack([((sensitivities >= x).sum(1)>=y).sum() 
+                for x in sens_fractions
+                for y in sens_counters]).reshape(
+                    sens_fractions.shape[0], sens_counters.shape[0])
+        return (sens_fractions, sens_counters, sensmat / self.aks.shape[0])
+
+    def sensitivity_range_matrix_z(self, z,
+            arr_dz = numpy.linspace(0, 30e3, 11),
+            arr_sens = numpy.linspace(0, 1, 11)):
+        """Like sensitivity_range_matrix but with elevation units.
+
+        Returns a matrix with the fraction of elements where sensitivity
+        exceeds 'x' for a range of elevations 'dz'.
+
+        :param z: Matrix containing elevations.  Shape must match
+            self.aks.shape[1:].
+        :param arr_dz: Array of delta-z to consider.
+        :param arr_sens: Array of sensitivities to consider.
+        :returns: (arr_dz, arr_sens, mat)
+        """
+
+        mat = numpy.zeros(shape=(arr_dz.size, arr_sens.size))
+        sensitivities = self.sensitivities()
+        with numpy.errstate(invalid="ignore"):
+            allmsk = [sensitivities > x for x in arr_sens]
+        # find highest and lowest z for each column, corresponding to msk
+        #
+        # Note: in some cases there are 'gaps', i.e. sensitivity mask
+        # looks like [True, True, False, False, False, True, True, True,
+        # True, True, True, True, True, False, False, ...].  In this case
+        # we take the 'False' along for now.  This may also yield more
+        # than one 'last'.
+
+        for (msk_i, msk) in enumerate(allmsk):
+            # first True in each column
+            mskno = msk.nonzero()
+            if not msk.any(): # all mat 0
+                continue
+            (_, ii) = numpy.unique(mskno[0], return_index=True) 
+            makes_sense = msk.any(1)
+            firsts = numpy.zeros(shape=sensitivities.shape[0])
+            firsts[makes_sense] = mskno[1][ii] # NB: goes wrong if 0 True values
+            firsts[~makes_sense] = -1
+
+            # last True in each column is element before!
+            lasts = numpy.zeros(shape=sensitivities.shape[0])
+            lasts[makes_sense] = numpy.hstack((mskno[1][ii[1:]-1], mskno[1][-1]))
+            lasts[~makes_sense] = 0
+
+            lower = numpy.array([z[i, firsts[i]] for i in range(firsts.shape[0])])
+            upper = numpy.array([z[i, lasts[i]] for i in range(lasts.shape[0])])
+            dz = upper - lower
+            # by handling > as false, nans are counted as not in range.
+            # There may be nans in z
+            with numpy.errstate(invalid="ignore"):
+                for (dz_k, dz_lim) in enumerate(arr_dz):
+                    mat[dz_k, msk_i] = (dz >= dz_lim).sum() / sensitivities.shape[0]
+
+        return (arr_dz, arr_sens, mat)
+
+
+    def plot_sensitivity_density(self,
+            nstep=11,
+            z=None):
+        """Visualise where sensors are typically sensitive
+        """
+
+        (sens_frac, sensmat) = self.sensitivity_density_matrix()
+
+        # regrid sensitivity matrix for z
+        if z.ndim > 1:
+            if (z.min(0) == z.max(0)).all():
+                z = z[0, :]
+            else:
+                newz = numpy.nanmean(z, 0)
+                logging.info("Regridding sensitivity matrices")
+                A_new = pamath.regrid_matrix(sensmat, z, newz)
+                logging.info("Done")
+                z = newz
+
+        # write some diagnostics
+        for (i, f) in enumerate(sens_frac):
+            for p in (0.2, 0.5, 0.8):
+                makes_sense = z[sensmat[i, :]>p]
+                if makes_sense.any():
+                    logging.info(("Altitude range with at least "
+                        "{:.0%} >{:.0%} sensitive: {:.1f}--{:.1f} km").format(
+                            p, f, makes_sense.min()/1e3,
+                            makes_sense.max()/1e3))
+                else:
+                    logging.info(("Never more than {:.0%} with "
+                        "sensitivity {:.0%} :(").format(
+                        sensmat[i, :].max(), f))
+            
+        #f = matplotlib.pyplot.figure()
+        (f, a) = matplotlib.pyplot.subplots() # = f.add_subplot(1, 1, 1)
+        cs = a.contourf(sens_frac, z, sensmat.T,
+            numpy.linspace(0, 1, nstep),
+            cmap=self.cmap)
+        #cs.clabel(colors="blue")
+        cb = f.colorbar(cs)
+        a.set_xlabel("Sensitivity")
+        a.set_ylabel("Elevation [m]")
+        cb.set_label("Fraction")
+        a.set_title("Elevation sensitivity density {}".format(self.name))
+        a.grid(which="major", color="white")
+        
+        graphics.print_or_show(
+            f, False, self.filename.format(mode="density_z", name=self.name))
+
+    def plot_sensitivity_range(self,
+            nstep=11,
+            z=None):
+        """Visualise vertical range of sensitivities
+        """
+
+        # Degrees of freedom according to:
+        #   Rodgers (2000)
+        #   Equation (2.80), Page 37
+
+        dofs = self.dofs()
+
+        sensitivities = self.sensitivities()
+
+        max_sensitivities = sensitivities.max(1)
+
+        (sens_fractions, sens_counters, sensmat) = self.sensitivity_range_matrix()
+
+        f = matplotlib.pyplot.figure()
+        a = f.add_subplot(1, 1, 1)
+        cs = a.contourf(sens_fractions, sens_counters, sensmat.T,
+            numpy.linspace(0, 1, nstep),
+            cmap=self.cmap)
+        #cs.clabel(colors="blue")
+        cb = f.colorbar(cs)
+        a.set_xlabel("Sensitivity")
+        a.set_ylabel("No. of levels")
+        cb.set_label("Fraction")
+        a.set_title(("Fraction of profiles with at least N layers "
+                     "sensitivity > x"))
+        graphics.print_or_show(
+            f, False, self.filename.format(mode="range_n", name=self.name))
+
+        (arr_dz, arr_sens, mat_z) = self.sensitivity_range_matrix_z(z=z)
+
+        f = matplotlib.pyplot.figure()
+        a = f.add_subplot(1, 1, 1)
+        cs = a.contourf(arr_sens, arr_dz, mat_z,
+            numpy.linspace(0, 1, nstep),
+            cmap=self.cmap)
+        #cs.clabel(colors="blue")
+        cb = f.colorbar(cs)
+        a.set_xlabel("Sensitivity")
+        a.set_ylabel("Delta z [m]")
+        cb.set_label("Fraction")
+        a.set_title(("Fraction of profiles with sensitivity > x "
+                     "throughout a certain vertical range"))
+        graphics.print_or_show(
+            f, False, self.filename.format(mode="range_z", name=self.name))
+
 
 def mixingratio2density(mixingratio, p, T):
     """Converts mixing ratio (e.g. kg/kg) to density (kg/m^3) for dry air.
